#include <tcl.h>

#include <string.h>
#include <malloc.h>
#include "version.h"

#include "msql.h"

#define MAX_HOST_LEN 256


Tcl_CmdProc Msqltcl_msqlConnect;

static int msqlSocketCount = 0;
static Tcl_HashTable msqlSockets;

typedef struct _msqlSocketData
{
	char host[MAX_HOST_LEN];
	int socket;
} MsqlSocketData;

int Msqltcl_msqlConnect(
	ClientData	notUsed,
	Tcl_Interp*	interp,
	int			argc,
	char**		argv)
{
	char *host, socket_tag[256];
	int socket, create_flag;
	MsqlSocketData *msqlSocketData;
	Tcl_HashEntry *entry;

	if (argc > 2 ) {
		Tcl_AppendResult(interp, "usage : ", argv[0], " [<host>]", NULL);
		return TCL_ERROR;
	}

	if (argc == 2)
		host = argv[1];
	else
		host = NULL;

	socket = msqlConnect(host);
	if (socket == -1)
	{
		if (!host)
			host = "/dev/msql";

		Tcl_AppendResult(interp, "msqltcl: unable to connect to ", host, NULL);
		Tcl_AppendResult(interp, msqlErrMsg, NULL);
		return TCL_ERROR;
	}

	sprintf(socket_tag, "msql%d", msqlSocketCount);

	msqlSocketData = (MsqlSocketData *) ckalloc(sizeof(MsqlSocketData));
	if (host)
		strncpy(msqlSocketData->host, host, MAX_HOST_LEN);
	else
		msqlSocketData->host[0] = '\0';

	msqlSocketData->socket = socket;

	entry = Tcl_CreateHashEntry(&msqlSockets, socket_tag, &create_flag);
	Tcl_SetHashValue(entry, (ClientData) msqlSocketData);

	Tcl_SetResult(interp, socket_tag, TCL_VOLATILE);

	++msqlSocketCount;

	return TCL_OK;
}

int Msqltcl_msqlSelectDB(
	ClientData	notUsed,
	Tcl_Interp*	interp,
	int			argc,
	char**		argv)
{
	MsqlSocketData *msqlSocketData;
	Tcl_HashEntry *entry;
	int status;

	if (argc != 3)
	{
		Tcl_AppendResult(interp, "usage: ", *argv, " <handle> <dbName>", NULL);
		return TCL_ERROR;
	}

	entry = Tcl_FindHashEntry(&msqlSockets, argv[1]);
	if (!entry)
	{
		Tcl_AppendResult(interp, *argv, ": bad handle \"", argv[1], "\"", NULL);
		return TCL_ERROR;
	}

	msqlSocketData = (MsqlSocketData *) Tcl_GetHashValue(entry);
	status = msqlSelectDB(msqlSocketData->socket, argv[2]);
	if (status == -1)
	{
		Tcl_AppendResult(interp,
			*argv, ": error selecting \"", argv[2],
			"\" for \"", argv[1], "\"\n",
			msqlErrMsg, NULL);
		return TCL_ERROR;
	}

	return TCL_OK;
}

int Msqltcl_msqlQuery(
	ClientData	notUsed,
	Tcl_Interp*	interp,
	int			argc,
	char**		argv)
{
	MsqlSocketData *msqlSocketData;
	Tcl_HashEntry *entry;
	int status, i, j;
	m_result *msql_result;
	Tcl_DString resultList;
	char *colmask, *command, *startPtr, *scanPtr;

	command = NULL;
	colmask = "%D";	/* D = Data */
	while (argc > 3)
	{
		if (strcmp(argv[argc - 2], "-command") == 0)
			command = argv[argc - 1];
		else if (strcmp(argv[argc - 2], "-colmask") == 0)
			colmask = argv[argc - 1];
		argc -=2;
	}

	if (argc != 3 && argc != 5)
	{
		Tcl_AppendResult(interp,
			"usage: ", *argv,
			" <handle> <query> [-command <command>] -colmask <colmask>",
			NULL);
		return TCL_ERROR;
	}

	entry = Tcl_FindHashEntry(&msqlSockets, argv[1]);
	if (!entry)
	{
		Tcl_AppendResult(interp, *argv, ": bad handle \"", argv[1], "\"", NULL);
		return TCL_ERROR;
	}

	msqlSocketData = (MsqlSocketData *) Tcl_GetHashValue(entry);

	status = msqlQuery(msqlSocketData->socket, argv[2]);
	if (status == -1)
	{
		Tcl_AppendResult(interp,
			*argv, ": error sending query \"", argv[2],
			"\" for \"", argv[1], "\"\n",
			msqlErrMsg, NULL);
		return TCL_ERROR;
	}

	Tcl_DStringInit (&resultList);


	msql_result = msqlStoreResult();
	if (!msql_result)
		return TCL_OK;

	for (i = 0; i < msqlNumRows(msql_result); i++)
	{
		m_row msql_row;

		msqlDataSeek(msql_result, i);
		msql_row = msqlFetchRow(msql_result);

		Tcl_DStringStartSublist(&resultList);

		for (j = 0; j < msqlNumFields(msql_result); j++)
		{
			m_field *msql_field;
			char num_buf[BUFSIZ], *result;

			msqlFieldSeek(msql_result, j);
			msql_field = msqlFetchField(msql_result);

/*			Tcl_DStringStartSublist(&resultList); */
			startPtr = colmask;
			for (scanPtr = startPtr; *scanPtr != '\0'; scanPtr++)
			{
				char tmp_buf[BUFSIZ];

				if (*scanPtr != '%')
					continue;
				Tcl_DStringAppend (&resultList, startPtr, scanPtr - startPtr);
				scanPtr++;
				startPtr = scanPtr + 1;
				switch (*scanPtr)
				{
				case 'D':
					Tcl_DStringAppendElement(&resultList, msql_row[j]);
					break;
				case 'N':
					Tcl_DStringAppendElement(&resultList, msql_field->name);
					break;
				case 'T':
					Tcl_DStringAppendElement(&resultList, msql_field->table);
					break;
				case 'C':
					switch (msql_field->type)
					{
					case INT_TYPE:
						Tcl_DStringAppendElement(&resultList, "INT");
						break;
					case REAL_TYPE:
						Tcl_DStringAppendElement(&resultList, "REAL");
						break;
					case CHAR_TYPE:
						Tcl_DStringAppendElement(&resultList, "CHAR");
						break;
					default:
						Tcl_DStringAppendElement(&resultList, "ERROR");
						break;
					}
					break;
				case 'L':
					Tcl_DStringAppendElement(&resultList,
						sprintf(tmp_buf, "%d", msql_field->length));
					break;
				case 'F':
					Tcl_DStringStartSublist(&resultList);
					if (IS_PRI_KEY(msql_field->flags))
						Tcl_DStringAppendElement(&resultList, "PRIMARY");
					if (IS_NOT_NULL(msql_field->flags))
						Tcl_DStringAppendElement(&resultList, "NOT_NULL");
					break;
					Tcl_DStringEndSublist(&resultList);
				default:
					Tcl_DStringAppendElement(&resultList, "ERROR");
				}
			}
/*			Tcl_DStringEndSublist(&resultList); */
		}
		Tcl_DStringEndSublist(&resultList);

		if (command)
		{
			Tcl_DString commandList;
			int result;

			Tcl_DStringInit(&commandList);
			startPtr = command;
			for (scanPtr = startPtr; *scanPtr != '\0'; scanPtr++)
			{
				char tmp_buf[BUFSIZ];

				if (*scanPtr != '%')
					continue;
				Tcl_DStringAppend (&commandList, startPtr, scanPtr - startPtr);
				scanPtr++;
				startPtr = scanPtr + 1;
				switch (*scanPtr)
				{
				case 'R':
					Tcl_DStringAppend(&commandList, resultList.string, -1);
					break;
				default:
					Tcl_DStringAppendElement(&commandList, "ERROR");
					break;
				}
			}
			Tcl_DStringAppend (&commandList, startPtr, scanPtr - startPtr);

			result = Tcl_GlobalEval(interp, commandList.string);
			if (result != TCL_OK)
			{
				Tcl_AppendResult(interp,
					*argv, " error avaluating command \"",
					commandList.string, "\"",
					NULL);
				Tcl_DStringFree(&commandList);
				Tcl_DStringFree(&resultList);
				msqlFreeResult(msql_result);
				return TCL_ERROR;
			}
			Tcl_DStringFree(&commandList);
			Tcl_DStringFree(&resultList);
			Tcl_DStringInit(&resultList);
		}
	}

	if (!command)
		Tcl_DStringResult(interp, &resultList);

	msqlFreeResult(msql_result);

	return TCL_OK;
}

int Msqltcl_msqlClose(
	ClientData	notUsed,
	Tcl_Interp*	interp,
	int			argc,
	char**		argv)
{
	MsqlSocketData *msqlSocketData;
	Tcl_HashEntry *entry;

	if (argc != 2)
	{
		Tcl_AppendResult(interp, "usage: ", *argv, " <handle>", NULL);
		return TCL_ERROR;
	}

	entry = Tcl_FindHashEntry(&msqlSockets, argv[1]);
	if (!entry)
	{
		Tcl_AppendResult(interp, *argv, ": bad handle \"", argv[1], "\"", NULL);
		return TCL_ERROR;
	}

	msqlSocketData = (MsqlSocketData *) Tcl_GetHashValue(entry);
	msqlClose(msqlSocketData->socket);
	Tcl_DeleteHashEntry(entry);
	ckfree((char *) msqlSocketData);

	return TCL_OK;
}

void Msqltcl_msqlInit()
{
	Tcl_InitHashTable(&msqlSockets, TCL_STRING_KEYS);
}

/*
 * This is where we tell the loadable module what the names of our functions
 * are.
 *
 * The name of this function is the name of the library.
 * eg;
 *      libdemo.so.1.1 -> Msqltcl_Init
 *      libngb.so.1.1 -> Ngb_Init
 */
int Msqltcl_Init(interp)
	Tcl_Interp *interp;
{
	Tcl_SetVar(interp, "demo_version", Version, TCL_GLOBAL_ONLY);

	Msqltcl_msqlInit();

	Tcl_CreateCommand(interp, "msqlConnect", Msqltcl_msqlConnect,
			(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand(interp, "msqlClose", Msqltcl_msqlClose,
			(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand(interp, "msqlSelectDB", Msqltcl_msqlSelectDB,
			(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateCommand(interp, "msqlQuery", Msqltcl_msqlQuery,
			(ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	return TCL_OK;
}

